import socket
import unittest
from contextlib import contextmanager
from logging import getLogger
from subprocess import run
from time import sleep, monotonic
from typing import Generator
from uuid import uuid4

from paramiko import SSHClient, AutoAddPolicy

from run_with_logger import run_with_logger__ssh


class TestRunWithLoggerSsh(unittest.TestCase):
    def test_paramiko__capture_stdout(self) -> None:
        logger = getLogger(__name__)

        port = 54321
        username = "user"
        password = "password"

        with ssh_server(
            port=port,
            username=username,
            password=password,
        ):
            ssh_client = SSHClient()
            ssh_client.set_missing_host_key_policy(policy=AutoAddPolicy)
            ssh_client.connect(
                hostname="localhost",
                port=port,
                username=username,
                password=password,
            )
            completed = run_with_logger__ssh(
                logger=logger,
                client=ssh_client,
                command="whoami",
                stdout_action="capture",
                stderr_action="capture",
            )

            self.assertEqual(0, completed.returncode)
            self.assertEqual(username, completed.stdout.decode().strip())
            self.assertEqual("", completed.stderr.decode().strip())

    def test_paramiko__capture_stderr(self) -> None:
        logger = getLogger(__name__)

        port = 54321
        username = "user"
        password = "password"

        with ssh_server(
            port=port,
            username=username,
            password=password,
        ):
            ssh_client = SSHClient()
            ssh_client.set_missing_host_key_policy(policy=AutoAddPolicy)
            ssh_client.connect(
                hostname="localhost",
                port=port,
                username=username,
                password=password,
            )
            completed = run_with_logger__ssh(
                logger=logger,
                client=ssh_client,
                command="whoami >&2",
                stdout_action="capture",
                stderr_action="capture",
            )

            self.assertEqual(0, completed.returncode)
            self.assertEqual("", completed.stdout.decode().strip())
            self.assertEqual(username, completed.stderr.decode().strip())

    def test_fabric__capture_stdout(self) -> None:
        from fabric import Connection

        logger = getLogger(__name__)

        port = 54321
        username = "user"
        password = "password"

        with ssh_server(
            port=port,
            username=username,
            password=password,
        ):
            with Connection(
                host="localhost",
                port=port,
                user=username,
                connect_kwargs={
                    "password": password,
                },
            ) as conn:
                completed = run_with_logger__ssh(
                    logger=logger,
                    client=conn,
                    command="whoami",
                    stdout_action="capture",
                    stderr_action="capture",
                )

            self.assertEqual(0, completed.returncode)
            self.assertEqual(username, completed.stdout.decode().strip())
            self.assertEqual("", completed.stderr.decode().strip())

    def test_fabric__capture_stderr(self) -> None:
        from fabric import Connection

        logger = getLogger(__name__)

        port = 54321
        username = "user"
        password = "password"

        with ssh_server(
            port=port,
            username=username,
            password=password,
        ):
            with Connection(
                host="localhost",
                port=port,
                user=username,
                connect_kwargs={
                    "password": password,
                },
            ) as conn:
                completed = run_with_logger__ssh(
                    logger=logger,
                    client=conn,
                    command="whoami >&2",
                    stdout_action="capture",
                    stderr_action="capture",
                )

            self.assertEqual(0, completed.returncode)
            self.assertEqual("", completed.stdout.decode().strip())
            self.assertEqual(username, completed.stderr.decode().strip())


@contextmanager
def ssh_server(
    *,
    port: int,
    username: str,
    password: str,
) -> Generator[None, None, None]:
    """
    Spin up a temporary SSH server for testing.
    """
    container_name = f"ssh-server-{uuid4()}"
    try:
        run(
            args=[
                "docker",
                "container",
                "run",
                "--name",
                container_name,
                "--rm",
                "-d",
                "-p",
                f"{port}:2222",
                "-e",
                f"USER_NAME={username}",
                "-e",
                f"USER_PASSWORD={password}",
                "-e",
                f"PASSWORD_ACCESS=true",
                "linuxserver/openssh-server",
            ],
            check=True,
            capture_output=True,
        )

        # Poll until the port is open.
        wait_for_ssh_server(host="localhost", port=port, timeout=10)

        yield

    finally:
        stop_and_remove_container(name=container_name)


def wait_for_ssh_server(*, host: str, port: int, timeout: float) -> None:
    deadline = monotonic() + timeout
    while True:
        try:
            with socket.create_connection((host, port), timeout=1) as sock:
                # Read the SSH banner to verify the server is ready
                sock.settimeout(1)
                banner = sock.recv(1024).decode("utf-8", errors="ignore").strip()
                if banner.startswith("SSH-"):
                    # Valid SSH banner received
                    break
                else:
                    raise ValueError(f"Invalid SSH banner: {banner}")
        except (OSError, ValueError):
            if monotonic() >= deadline:
                raise RuntimeError(f"Timed out waiting for SSH server on {host}:{port}")
            sleep(0.1)


def stop_and_remove_container(*, name: str) -> None:
    logger = getLogger(__name__)

    completed = run(
        args=[
            "docker",
            "container",
            "stop",
            name,
        ],
        check=False,
        capture_output=True,
        text=True,
    )
    if completed.returncode:
        if "No such container" in completed.stderr:
            return

        logger.error(f"Failed to stop container `{name}`:\n{completed.stderr}")

    completed = run(
        args=[
            "docker",
            "container",
            "rm",
            name,
        ],
        check=False,
        capture_output=True,
        text=True,
    )
    if completed.returncode:
        if "No such container" in completed.stderr:
            return

        logger.error(f"Failed to remove container `{name}`:\n{completed.stderr}")



import json
from io import BytesIO
from logging import Logger, DEBUG
from os import PathLike, environ
from subprocess import CompletedProcess, Popen, DEVNULL, PIPE, CalledProcessError
from time import sleep
from typing import Optional, Union, IO

from ._pipe_capture import pipe_capture__thread
from ._pipe_to_logger import pipe_to_logger__thread
from ._types import PopenArgsT, PopenCwdT, StreamActionT


def run_with_logger(
    *,
    args: PopenArgsT,
    logger: Logger,
    cwd: PopenCwdT = None,
    shell: bool = False,
    level: int = DEBUG,
    encoding: Optional[str] = None,
    check: bool = True,
    stdout_action: StreamActionT = "log",
    stderr_action: StreamActionT = "log",
    stdin_data: Optional[bytes] = None,
    stdin_io: Union[None, int, IO[bytes]] = None,
    extra_env: Optional[dict[str, str]] = None,
) -> CompletedProcess[bytes]:
    """
    Like `subprocess.run`, but with the ability to pipe `stdout` and/or `stderr` to a `Logger` or capture each stream
    independently while the process is running.

    Args:
        args: The command line program and arguments.
        cwd: The working directory.
        shell: Whether to run the command in a shell.
        logger: The logger to which to pipe stdout and/or stderr.
        level: The logging level to use.
        encoding: The encoding to use for stdout and stderr when logging. Defaults to "utf-8".
        check: Whether to raise an exception if the process returns a non-zero exit code.
        stdout_action: What to do with stdout. See `StreamActionT`.
        stderr_action: What to do with stderr. See `StreamActionT`.
        stdin_data: Data to write to the stdin of the process.
        stdin_io: A file object to pass directly to Popen.
        extra_env:
            Extra environment variables to set for the process.
            These will be added to the current environment.
    """
    stdout_buffer: Optional[IO[bytes]] = None
    stderr_buffer: Optional[IO[bytes]] = None

    if stdin_io is not None and stdin_data is not None:
        raise ValueError("Only one of `stdin_io` or `stdin_data` may be specified.")

    if isinstance(args, (str, bytes, PathLike)):
        args_str = str(args)
    else:
        args_str = json.dumps([str(a) for a in args])
    logger.debug(f"Starting process: {args_str}")

    with Popen(
        cwd=cwd,
        args=args,
        shell=shell,
        stdout=DEVNULL if stdout_action == "discard" else PIPE,
        stderr=DEVNULL if stderr_action == "discard" else PIPE,
        stdin=PIPE if stdin_data is not None else stdin_io,
        env=None if extra_env is None else {**environ, **extra_env},
    ) as process:
        if process.stdin and stdin_data is not None:
            with process.stdin as f:
                f.write(stdin_data)

        if stdout_action == "log" and process.stdout:
            stdout_cm = pipe_to_logger__thread(
                pipe=process.stdout,
                logger=logger,
                level=level,
                encoding=encoding,
            )
        elif stdout_action == "capture" and process.stdout:
            stdout_buffer = BytesIO()
            stdout_cm = pipe_capture__thread(
                pipe=process.stdout,
                destination=stdout_buffer,
            )
        else:
            stdout_cm = nullcontext()  # type: ignore

        if stderr_action == "log" and process.stderr:
            stderr_cm = pipe_to_logger__thread(
                pipe=process.stderr,
                logger=logger,
                level=level,
                encoding=encoding,
            )
        elif stderr_action == "capture" and process.stderr:
            stderr_buffer = BytesIO()
            stderr_cm = pipe_capture__thread(
                pipe=process.stderr,
                destination=stderr_buffer,
            )
        else:
            stderr_cm = nullcontext()  # type: ignore

        with stdout_cm, stderr_cm:
            # According to https://docs.python.org/3/library/subprocess.html#subprocess.Popen.wait , `process.wait` can
            # cause a deadlock. So we simply poll the process until it is done, and sleep so that the threads can do
            # their work.
            while process.poll() is None:
                sleep(0.1)

            code = process.returncode

    stdout_result = stdout_buffer.getvalue() if stdout_buffer else None  # type: ignore
    stderr_result = stderr_buffer.getvalue() if stderr_buffer else None  # type: ignore

    if check and code:
        raise CalledProcessError(
            returncode=code,
            cmd=process.args,
            output=stdout_result,
            stderr=stderr_result,
        )

    return CompletedProcess(
        args=process.args,
        returncode=code,
        stdout=stdout_result,
        stderr=stderr_result,
    )



import json
import sys
import unittest
from logging import getLogger, DEBUG, INFO
from os import environ
from pathlib import Path

from run_with_logger import run_with_logger


class TestRunWithLogger(unittest.TestCase):
    maxDiff = None

    def test_log_stderr_and_capture_stdout(self) -> None:
        """
        This would cause a deadlock if `run_with_logger` used `process.wait` internally.
        See https://docs.python.org/3/library/subprocess.html#subprocess.Popen.wait
        """
        n = 10000  # The deadlock only happens for large values of `n`.
        logger = getLogger(__name__)
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            # This needs to be a program that produces a lot of output on both stdout and stderr.
            args = [
                sys.executable,
                "-c",
                f"""\
import sys
for i in range({n}):
    print(f"stdout {{i+1}}")
    print(f"stderr {{i+1}}", file=sys.stderr)
""",
            ]

            completed = run_with_logger(
                args=args,
                logger=logger,
                stdout_action="capture",
                stderr_action="log",
                check=False,
            )

        self.assertEqual(
            [f"Starting process: {json.dumps(args)}"]
            + [f"stderr {i+1}" for i in range(n)],
            [r.getMessage() for r in cm.records],
        )
        self.assertEqual(
            [f"stdout {i+1}" for i in range(n)],
            completed.stdout.decode().splitlines(),
        )
        self.assertEqual(0, completed.returncode)

    def test_log_capture_stderr_and_log_stdout(self) -> None:
        """
        This would cause a deadlock if `run_with_logger` used `process.wait` internally.
        See https://docs.python.org/3/library/subprocess.html#subprocess.Popen.wait
        """
        n = 10000  # The deadlock only happens for large values of `n`.
        logger = getLogger(__name__)
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            # This needs to be a program that produces a lot of output on both stdout and stderr.
            args = [
                sys.executable,
                "-c",
                f"""\
import sys
for i in range({n}):
    print(f"stdout {{i+1}}")
    print(f"stderr {{i+1}}", file=sys.stderr)
""",
            ]

            completed = run_with_logger(
                args=args,
                logger=logger,
                stdout_action="log",
                stderr_action="capture",
                check=False,
            )

        self.assertEqual(
            [f"Starting process: {json.dumps(args)}"]
            + [f"stdout {i+1}" for i in range(n)],
            [r.getMessage() for r in cm.records],
        )
        self.assertEqual(
            [f"stderr {i+1}" for i in range(n)],
            completed.stderr.decode().splitlines(),
        )
        self.assertEqual(0, completed.returncode)

    def test_path_arguments(self) -> None:
        """
        This tests that `run_with_logger` can handle `Path` objects in the `args` list.
        """

        logger = getLogger(__name__)
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            completed = run_with_logger(
                logger=logger,
                level=INFO,
                args=[
                    sys.executable,
                    "-c",
                    "import sys; print(sys.argv[1])",
                    Path("/foo/bar"),
                ],
            )

        self.assertEqual(
            [
                (
                    DEBUG,
                    f'Starting process: ["{sys.executable}", "-c", "import sys; print(sys.argv[1])", "/foo/bar"]',
                ),
                (
                    INFO,
                    "/foo/bar",
                ),
            ],
            [(r.levelno, r.getMessage()) for r in cm.records],
        )
        self.assertEqual(0, completed.returncode)

    def test_shell_string(self) -> None:
        """
        Test with a string passed to a shell.
        """
        logger = getLogger(__name__)
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            completed = run_with_logger(
                logger=logger,
                level=INFO,
                args=f"{sys.executable} -c 'import sys; print(sys.argv[1])' /foo/bar",
                shell=True,
            )

        self.assertEqual(
            [
                (
                    DEBUG,
                    f"Starting process: {sys.executable} -c 'import sys; print(sys.argv[1])' /foo/bar",
                ),
                (
                    INFO,
                    "/foo/bar",
                ),
            ],
            [(r.levelno, r.getMessage()) for r in cm.records],
        )
        self.assertEqual(0, completed.returncode)

    def test_shell_path(self) -> None:
        """
        Test with a Path passed to a shell.
        """
        logger = getLogger(__name__)
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            with self.assertRaises(TypeError):
                # TypeError: path-like args is not allowed when shell is true
                run_with_logger(
                    logger=logger,
                    level=INFO,
                    args=Path(sys.executable),
                    shell=True,
                )

        self.assertEqual(
            [(DEBUG, f"Starting process: {sys.executable}")],
            [(r.levelno, r.getMessage()) for r in cm.records],
        )

    def test_extra_env(self) -> None:
        """
        Test that `extra_env` is passed correctly.
        """
        logger = getLogger(__name__)

        py_code = "import os; import json; print(json.dumps(dict(os.environ)))"

        with self.assertNoLogs(logger=logger):
            completed = run_with_logger(
                logger=logger,
                level=INFO,
                args=[sys.executable, "-c", py_code],
                extra_env=None,
                stdout_action="capture",
            )
            baseline = json.loads(completed.stdout)

        with self.assertNoLogs(logger=logger):
            completed = run_with_logger(
                logger=logger,
                level=INFO,
                args=[sys.executable, "-c", py_code],
                extra_env={},
                stdout_action="capture",
            )
            empty_extras = json.loads(completed.stdout)

        with self.assertNoLogs(logger=logger):
            completed = run_with_logger(
                logger=logger,
                level=INFO,
                args=[sys.executable, "-c", py_code],
                extra_env={
                    "FOO": "BAR",
                    "BAZ": "QUX",
                },
                stdout_action="capture",
            )
            with_extras = json.loads(completed.stdout)

        self.assertTrue(len(environ) > 0)
        self.assertEqual(baseline.keys(), environ.keys())
        self.assertEqual(empty_extras.keys(), environ.keys())
        self.assertEqual({"BAZ", "FOO"}, with_extras.keys() - environ.keys())



from contextlib import contextmanager
from logging import Logger
from threading import Thread
from typing import Union, IO, Optional, Generator


def pipe_to_logger(
    *,
    pipe: Union[IO[bytes], IO[str]],
    logger: Logger,
    level: int,
    encoding: Optional[str] = None,
) -> None:
    """
    Pipe the STDERR or STDOUT of a Process to a logger.

    Run this function in a separate thread.

    See https://stackoverflow.com/questions/35488927/send-subprocess-popen-stdout-stderr-to-logging-module

    Args:
        pipe: The STDERR or STDOUT pipe of a subprocess.
        logger: The logger to which to pipe the output line by line.
        level: The logging level to use.
        encoding:
            An optional encoding to use when decoding bytes to strings. Defaults to UTF-8. Ignored if the pipe is a
            text pipe.
    """
    encoding = encoding or "utf-8"

    with pipe:
        for line in pipe:
            if isinstance(line, bytes):
                line = line.decode(
                    encoding=encoding,
                    errors="replace",
                ).strip()
            logger.log(level=level, msg=line)


@contextmanager
def pipe_to_logger__thread(
    *,
    pipe: Union[IO[bytes], IO[str], None],
    logger: Logger,
    level: int,
    encoding: Optional[str] = None,
) -> Generator[None, None, None]:
    """
    Context manager that runs a separate thread to pipe the STDERR or STDOUT of a Process to a logger.

    See https://stackoverflow.com/questions/35488927/send-subprocess-popen-stdout-stderr-to-logging-module

    Args:
        pipe: The STDERR or STDOUT pipe of a subprocess.
        logger: The logger to which to pipe the output line by line.
        level: The logging level to use.
        encoding:
            An optional encoding to use when decoding bytes to strings. Defaults to UTF-8. Ignored if the pipe is a
            text pipe.
    """
    if not pipe:
        # This context manager becomes a dummy.
        yield
        return

    thread = Thread(
        target=pipe_to_logger,
        kwargs=dict(
            pipe=pipe,
            logger=logger,
            level=level,
            encoding=encoding,
        ),
    )
    try:
        thread.start()
        yield
    finally:
        thread.join()



from io import BytesIO
from logging import Logger, DEBUG
from subprocess import CompletedProcess, CalledProcessError
from typing import Optional, TYPE_CHECKING

from ._pipe_capture import pipe_capture__thread
from ._pipe_to_logger import pipe_to_logger__thread
from ._types import StreamActionT

if TYPE_CHECKING:
    from paramiko import SSHClient
    from fabric import Connection


def run_with_logger__ssh(
    *,
    client: "SSHClient | Connection",
    command: str,
    logger: Logger,
    level: int = DEBUG,
    encoding: Optional[str] = None,
    check: bool = True,
    stdout_action: StreamActionT = "log",
    stderr_action: StreamActionT = "log",
    stdin_data: Optional[bytes] = None,
    extra_env: Optional[dict[str, str]] = None,
) -> CompletedProcess[bytes]:
    """
    Like `run_with_logger`, but via SSH.

    Args:
        client: The `paramiko` SSH client to use for running the command.
        command: The command line program and arguments.
        logger: The logger to which to pipe stdout and/or stderr.
        level: The logging level to use.
        encoding: The encoding to use for stdout and stderr when logging. Defaults to "utf-8".
        check: Whether to raise an exception if the process returns a non-zero exit code.
        stdout_action: What to do with stdout. See `StreamActionT`.
        stderr_action: What to do with stderr. See `StreamActionT`.
        stdin_data: Data to write to the stdin of the process.
        extra_env:
            Extra environment variables to set for the process.
            These will be added to the current environment.
    """
    from paramiko import SSHClient

    if not isinstance(client, SSHClient):
        from fabric import Connection

        if isinstance(client, Connection):
            # Open the connection and get the underlying Paramiko client.
            client.open()
            client = client.client

        else:
            raise ValueError(
                "`client` must be either a `paramiko.SSHClient` or a `fabric.Connection`."
            )

    if client.get_transport() is None or not client.get_transport().is_active():
        raise ValueError("SSH client is not connected.")

    logger.debug(f"Starting remote process: {command}")

    stdin_stream, stdout_stream, stderr_stream = client.exec_command(
        command=command,
        environment=extra_env,
    )

    # Handle stdout
    stdout_buffer: Optional[BytesIO] = None
    if stdout_action == "log":
        pipe_to_logger__thread(
            pipe=stdout_stream,
            logger=logger,
            level=level,
            encoding=encoding,
        )
    elif stdout_action == "capture":
        stdout_buffer = BytesIO()
        pipe_capture__thread(
            pipe=stdout_stream,
            destination=stdout_buffer,
        )

    # Handle stderr
    stderr_buffer: Optional[BytesIO] = None
    if stderr_action == "log":
        pipe_to_logger__thread(
            pipe=stderr_stream,
            logger=logger,
            level=level,
            encoding=encoding,
        )
    elif stderr_action == "capture":
        stderr_buffer = BytesIO()
        pipe_capture__thread(
            pipe=stderr_stream,
            destination=stderr_buffer,
        )

    # Write stdin data if provided
    if stdin_data is not None:
        stdin_stream.write(stdin_data)
        stdin_stream.close()

    # Wait for command completion
    code = stdout_stream.channel.recv_exit_status()

    if stdout_buffer:
        # Capture the remaining stdout data after the command has completed
        stdout_buffer.write(stdout_stream.read())
        stdout_result = stdout_buffer.getvalue()
    else:
        stdout_result = None

    if stderr_buffer:
        # Capture the remaining stderr data after the command has completed
        stderr_buffer.write(stderr_stream.read())
        stderr_result = stderr_buffer.getvalue()
    else:
        stderr_result = None

    if check and code:
        raise CalledProcessError(
            returncode=code,
            cmd=command,
            output=stdout_result,
            stderr=stderr_result,
        )

    return CompletedProcess(
        args=command,
        returncode=code,
        stdout=stdout_result,
        stderr=stderr_result,
    )



import doctest
import unittest
from pathlib import Path
from typing import Any, List
from unittest import BaseTestSuite

repo_dir = Path(__file__).parent.parent.resolve()


# noinspection PyUnusedLocal
def load_tests(loader: Any, tests: BaseTestSuite, ignore: Any) -> BaseTestSuite:
    """
    See https://docs.python.org/3/library/doctest.html#unittest-api
    """
    modules = find_modules_with_doctests()
    for module in modules:
        tests.addTests(doctest.DocTestSuite(module))
    return tests


def find_modules_with_doctests() -> List[str]:
    modules = []
    skip_n_parts = len(repo_dir.parts)
    for path in repo_dir.joinpath("src").rglob("*.py"):
        if path.name == "__init__.py":
            continue

        module = ".".join(path.parts[skip_n_parts:])
        module = module[:-3]
        modules.append(module)
    return modules


if __name__ == "__main__":
    unittest.main(
        failfast=True,
    )



from contextlib import contextmanager
from threading import Thread
from typing import IO, Generator


def pipe_capture(
    *,
    pipe: IO[bytes],
    destination: IO[bytes],
) -> None:
    """
    Capture the STDERR or STDOUT of a Process.

    Run this function in a separate thread.

    Args:
        pipe: The STDERR or STDOUT pipe of a subprocess.
        destination: The destination to which to pipe the output line by line.
    """
    with pipe:
        for line in pipe:
            destination.write(line)


@contextmanager
def pipe_capture__thread(
    *,
    pipe: IO[bytes],
    destination: IO[bytes],
) -> Generator[None, None, None]:
    """
    Context manager that runs a separate thread to capture the STDERR or STDOUT of a Process.

    Args:
        pipe: The STDERR or STDOUT pipe of a subprocess.
        destination: The destination to which to pipe the output line by line.
    """
    thread = Thread(
        target=pipe_capture,
        kwargs=dict(
            pipe=pipe,
            destination=destination,
        ),
    )
    try:
        thread.start()
        yield
    finally:
        thread.join()



from os import PathLike
from typing import Literal, Union, Sequence

PopenArgsT = Union[
    str,
    bytes,
    PathLike[str],
    PathLike[bytes],
    Sequence[
        Union[
            str,
            bytes,
            PathLike[str],
            PathLike[bytes],
        ]
    ],
]

PopenCwdT = Union[str, bytes, PathLike[str], PathLike[bytes], None]

StreamActionT = Literal["log", "capture", "discard"]
"""
What to do with a stream:
    - "log": Pipe to the logger.
    - "capture": Capture and return as part of `CompletedProcess`.
    - "discard": Ignore it (like piping it to `/dev/null`).
"""
